#!/usr/bin/env python3

import rospy  # Libreria principale per ROS (Robot Operating System)
import numpy as np  # Libreria per operazioni numeriche
from nav_msgs.msg import Odometry  # Messaggio ROS per dati di odometria
from geometry_msgs.msg import Twist, Point  # Messaggi ROS per comandi di velocità e punti
from tf import transformations  # Libreria per trasformazioni di coordinate

def control(x_ref_b, y_ref_b, x_ref_b_dot, y_ref_b_dot):
    """
    Funzione di controllo cinematico.
    Calcola le velocità lineari e angolari del robot per seguire una traiettoria di riferimento.

    Args:
        x_ref_b (float): Posizione di riferimento X del punto di controllo nel frame del robot.
        y_ref_b (float): Posizione di riferimento Y del punto di controllo nel frame del robot.
        x_ref_b_dot (float): Velocità di riferimento X del punto di controllo nel frame del robot.
        y_ref_b_dot (float): Velocità di riferimento Y del punto di controllo nel frame del robot.

    Returns:
        tuple: Velocità lineare e angolare calcolate per il robot.
    """
    global b, current_pose, kx, ky
    theta = current_pose[2]  # Orientamento attuale del robot
    xb = current_pose[0] + b * np.cos(theta)  # Posizione X del punto di controllo
    yb = current_pose[1] + b * np.sin(theta)  # Posizione Y del punto di controllo

    # Matrice Jacobiana: mette in relazione le velocità del robot con le velocità del punto di controllo
    D = np.array([
        [np.cos(theta), -b * np.sin(theta)],
        [np.sin(theta), b * np.cos(theta)]
                  ])

    # Inversa della Jacobiana: permette di calcolare le velocità del robot a partire dalle velocità del punto di controllo
    D_inv = np.linalg.inv(D)

    # Vettore di controllo: contiene gli errori di posizione e le velocità di riferimento
    control_matrix = np.array([
        [-kx * (xb - x_ref_b) + x_ref_b_dot],
        [-ky * (yb - y_ref_b) + y_ref_b_dot]
    ])

    # Calcolo delle velocità lineari e angolari del robot
    v_c, omega_c = np.dot(D_inv, control_matrix).flatten()

    return v_c, omega_c

def stop_node():
    """Arresta il robot pubblicando comandi di velocità nulli."""
    global cmd_vel_publisher, rate
    stop_msg = Twist()  # Crea un messaggio Twist con velocità nulle
    for step in range(10):
        cmd_vel_publisher.publish(stop_msg)  # Pubblica il messaggio sulla velocità comando
        rate.sleep()  # Mantiene la frequenza del loop

def pose_updater(msg):
    """Aggiorna la posa del robot con i dati di odometria ricevuti."""
    global current_pose
    x = msg.pose.pose.position.x  # Estrae la posizione x
    y = msg.pose.pose.position.y  # Estrae la posizione y
    orientation = np.array([msg.pose.pose.orientation.x,
                            msg.pose.pose.orientation.y,
                            msg.pose.pose.orientation.z,
                            msg.pose.pose.orientation.w])  # Estrae l'orientamento (quaternione)
    _, _, theta = transformations.euler_from_quaternion(orientation)  # Converte il quaternione in angoli di Eulero (theta)
    current_pose = np.array([x, y, theta])  # Aggiorna la variabile globale current_pose

def compute_reference_trajectory():
    """
    Funzione per generare la traiettoria di riferimento.
    """
    global trajectory_type, rotation_matrix, init_x, init_y, init_theta, set_point_list, set_point_dot_list
    if trajectory_type == "retta":
        vel = 0.08
        d = 2.0
        traj = np.arange(start=0., stop=d, step=vel / hz)
        for i in range(len(traj)):
            point = np.array([traj[i], 0.]).T
            point_dot = np.array([vel, 0]).T
            point_w = np.dot(rotation_matrix, point).reshape((2,)) + np.array([init_x, init_y])
            point_dot_w = np.dot(rotation_matrix, point_dot).reshape((2,))
            set_point_list.append(point_w)
            set_point_dot_list.append(point_dot_w)

    elif trajectory_type == "sinusoide":
        vel = 0.08
        d = 3.0
        time_steps = np.arange(start=0., stop=d / vel, step=1 / hz)

        r = 0.4
        w = vel / r

        y_coord = r * np.sin(w * time_steps)
        x_coord = time_steps * vel * 0.
        y_coord_dot = w * r * np.cos(w * time_steps)
        x_coord_dot = np.ones_like(y_coord_dot) * vel * 0.
        for i in range(len(y_coord)):
            point = np.array([x_coord[i], y_coord[i]]).T
            point_dot = np.array([x_coord_dot[i], y_coord_dot[i]]).T
            point_w = np.dot(rotation_matrix, point).reshape((2,)) + np.array([init_x, init_y])
            point_dot_w = np.dot(rotation_matrix, point_dot).reshape((2,))
            set_point_list.append(point_w)
            set_point_dot_list.append(point_dot_w)
    elif trajectory_type == "cerchio":
        v = 0.08
        r = 0.40
        w = v / r
        T = (2 * np.pi) / w

        time_steps = np.arange(start=0., stop=T, step=1 / hz)
        circular_x = r * np.sin(w * time_steps)
        circular_y = r * np.cos(w * time_steps) - r
        circular_x_dot = w * r * np.cos(w * time_steps)
        circular_y_dot = -w * r * np.sin(w * time_steps)

        for i in range(len(circular_x)):
            point = np.array([circular_x[i], circular_y[i]]).T
            point_dot = np.array([circular_x_dot[i], circular_y_dot[i]]).T
            point_w = np.dot(rotation_matrix, point).reshape((2,)) + np.array([init_x, init_y])
            point_dot_w = np.dot(rotation_matrix, point_dot).reshape((2,))
            set_point_list.append(point_w)
            set_point_dot_list.append(point_dot_w)
    else:
        print("Trajectory not supported")
        exit(-1)

    return set_point_list, set_point_dot_list


if __name__ == "__main__":
    rospy.init_node("simple_controller", anonymous=False)  # Inizializza il nodo ROS

    print("Init done")
    hz = 10  # Frequenza del loop di controllo
    rate = rospy.Rate(hz)  # Crea un oggetto Rate per mantenere la frequenza

    # Parametri del robot
    b = 0.1
    kx = 1
    ky = 1

    # Feedback Data
    current_pose = np.zeros((3,))
    pose_subscriber = rospy.Subscriber("/odom", Odometry, pose_updater, queue_size=10) # Crea un subscriber per il topic /odom

    # Robot Control
    cmd_vel_publisher = rospy.Publisher("/cmd_vel", Twist, queue_size=10) # Crea un publisher per il topic /cmd_vel

    # Ref Signal
    ref_signal_publisher = rospy.Publisher("/ref_signal", Point, queue_size=10) # Crea un publisher per il topic /ref_signal
    xb_signal_publisher = rospy.Publisher("/xb_signal", Point, queue_size=10) # Crea un publisher per il topic /xb_signal

    while not rospy.is_shutdown(): # Loop principale del nodo

        # Ref Signal Parameters
        trajectory_type = input("Insert Trajectory type [retta, sinusoide, cerchio]: ") # Ottiene il tipo di traiettoria dall'utente.

        init_msg = rospy.wait_for_message("/odom", Odometry) # Attende il primo messaggio di odometria
        init_orientation = np.array([init_msg.pose.pose.orientation.x,
                                     init_msg.pose.pose.orientation.y,
                                     init_msg.pose.pose.orientation.z,
                                     init_msg.pose.pose.orientation.w])
        _, _, init_theta = transformations.euler_from_quaternion(init_orientation)
        init_x = init_msg.pose.pose.position.x + b * np.cos(init_theta)
        init_y = init_msg.pose.pose.position.y + b * np.sin(init_theta)
        rotation_matrix = np.array([[np.cos(init_theta), -np.sin(init_theta)],
                                    [np.sin(init_theta), np.cos(init_theta)]])
        set_point_list = []
        set_point_dot_list = []
        compute_reference_trajectory() # Calcola la traiettoria di riferimento

        # Exit policy
        rospy.on_shutdown(stop_node) # Chiama la funzione stop_node quando il nodo viene chiuso

        # Control loop
        for set_point, set_point_dot in zip(set_point_list, set_point_dot_list): # Itera sui punti della traiettoria
            cmd_vel_msg = Twist() # Crea un messaggio Twist per il comando di velocità
            v, omega = control(set_point[0], set_point[1],
                               set_point_dot[0], set_point_dot[1]) # Calcola le velocità di controllo
            cmd_vel_msg.linear.x = v # Imposta la velocità lineare
            cmd_vel_msg.angular.z = omega # Imposta la velocità angolare
            cmd_vel_publisher.publish(cmd_vel_msg) # Pubblica il comando di velocità

            # Ref signal
            current_ref = Point() # Crea un messaggio Point per la posizione di riferimento
            current_ref.x = set_point[0] # Imposta la coordinata x
            current_ref.y = set_point[1] # Imposta la coordinata y

            ref_signal_publisher.publish(current_ref) # Pubblica la posizione di riferimento

            # Ref signal
            current_xb = Point() # Crea un messaggio Point per la posizione del punto di controllo
            current_xb.x = current_pose[0] + b * np.cos(current_pose[2]) # Calcola la coordinata x del punto di controllo
            current_xb.y = current_pose[1] + b * np.sin(current_pose[2]) # Calcola la coordinata y del punto di controllo

            xb_signal_publisher.publish(current_xb) # Pubblica la posizione del punto di controllo
            rate.sleep() # Mantiene la frequenza del loop.
